from __future__ import unicode_literals

from django.db import models


class AbstractTenantManager(models.Manager):
    def __init__(self):
        self.__cache = {}
        super(AbstractTenantManager, self).__init__()

    def clear_cache(self):
        for tenant in list(self.__cache.values()):
            self._remove_from_cache(tenant)

    def _get_from_cache(self, *natural_key):
        return self.__cache[natural_key]

    def _add_to_cache(self, tenant):
        key = tenant.natural_key()
        try:
            return self._get_from_cache(*key)
        except KeyError:
            self.__cache[key] = tenant
        return tenant

    def _remove_from_cache(self, tenant):
        key = tenant.natural_key()
        delattr(tenant, 'models')
        return self.__cache.pop(key)

    def _get_by_natural_key(self, *natural_key):
        raise NotImplementedError

    def get_by_natural_key(self, *natural_key):
        try:
            tenant = self._get_from_cache(*natural_key)
        except KeyError:
            tenant = self._add_to_cache(self._get_by_natural_key(*natural_key))
        return tenant


class TenantManager(AbstractTenantManager):
    def _get_by_natural_key(self, name):
        return self.get(name=name)


class TenantModelManagerDescriptor(object):
    """
    This class provides a better error message when you try to access a
    manager on an tenant model.
    """
    def __init__(self, model):
        self.model = model

    def __get__(self, instance, owner):
        raise AttributeError(
            "Manager isn't available; %s is tenant specific" % (
                self.model._meta.object_name,
            )
        )



from __future__ import unicode_literals

import logging

from django.dispatch.dispatcher import receiver
from django.utils import six
from django.utils.six.moves import copyreg
from mutant.db.models import MutableModel
from mutant.models import (
    BaseDefinition, ModelDefinition, OrderingFieldDefinition,
)
from mutant.models.model import MutableModelProxy
from mutant.signals import mutable_class_prepared

from .. import get_tenant_model
from ..compat import (
    get_remote_field, get_remote_field_model, set_remote_field_model,
)
from ..models import (
    Reference, TenantModel, TenantModelBase, TenantSpecificModel,
    db_schema_table,
)
from ..signals import (
    post_models_creation, pre_models_creation, pre_schema_deletion,
)
from ..utils import get_forward_fields, get_model, get_reverse_fields


class MutableReference(Reference):
    def for_tenant(self, tenant):
        return self.model.for_tenant(tenant)


class MutableTenantModelBase(TenantModelBase):
    reference = MutableReference

    @classmethod
    def tenant_model_bases(cls, tenant, bases):
        tenant_bases = super(MutableTenantModelBase, cls).tenant_model_bases(tenant, bases)
        return tuple(
            tenant_base.__get__(None, None)
            if isinstance(base, cls) and not base._meta.abstract else tenant_base
            for base, tenant_base in zip(bases, tenant_bases)
        )

    def for_tenant(self, tenant):
        if issubclass(self, TenantSpecificModel):
            raise ValueError('Can only be called on non-tenant specific model.')
        opts = self._meta
        if opts.proxy:
            return super(MutableTenantModelBase, self).for_tenant(tenant)

        reference = self.references[self]
        app_label = opts.app_label
        object_name = reference.object_name_for_tenant(tenant)

        # Return the already cached model instead of creating a new one.
        model = get_model(opts.app_label, object_name.lower())
        if model:
            return MutableModelProxy(model)

        base = self.abstract_tenant_model_factory(tenant)
        # Create the model definition as managed and unmanage it right after
        # to make sure tables are all created on tenant model creation.
        model_def, created = ModelDefinition.objects.get_or_create(
            app_label=app_label,
            object_name=object_name,
            defaults={
                'bases': (BaseDefinition(base=base),),
                'db_table': db_schema_table(tenant, opts.db_table),
                'managed': True
            }
        )
        if created:
            model_def.managed = False
            model_def.save(update_fields=['managed'])
            for order, lookup in enumerate(base._meta.ordering):
                if lookup.startswith('-'):
                    lookup = lookup[1:]
                    descending = True
                else:
                    descending = False
                OrderingFieldDefinition.objects.create(
                    model_def=model_def,
                    order=order,
                    lookup=lookup,
                    descending=descending
                )
        return model_def.model_class()


class MutableTenantModel(six.with_metaclass(MutableTenantModelBase, TenantModel)):
    class Meta:
        abstract = True

    @classmethod
    def get_model_state(cls, **kwargs):
        model_state = super(MutableTenantModel, cls).get_model_state(**kwargs)
        tenant_model = str(cls._for_tenant_model._meta)
        model_state.bases = tuple(
            base for base in model_state.bases if base != tenant_model
        )
        return model_state

    @classmethod
    def get_related_model_states(cls, model_state):
        model_states = super(MutableTenantModel, cls).get_related_model_states(model_state)
        apps = cls._meta.apps
        for state in model_states:
            model = apps.get_model(state.app_label, state.name)
            if issubclass(model, TenantSpecificModel):
                tenant_model = str(model._for_tenant_model._meta)
                state.bases = tuple(
                    base for base in state.bases if base != tenant_model
                )
        return model_states


def __unpickle_mutable_tenant_model_base(model, natural_key, abstract):
    try:
        manager = get_tenant_model()._default_manager
        tenant = manager.get_by_natural_key(*natural_key)
        if abstract:
            return model.abstract_tenant_model_factory(tenant)
        return model.for_tenant(tenant)
    except Exception:
        logger = logging.getLogger('tenancy.pickling')
        logger.exception('Failed to unpickle mutable tenant model')


def __pickle_mutable_tenant_model_base(model):
    if issubclass(model, TenantSpecificModel):
        tenant = getattr(model, get_tenant_model().ATTR_NAME)
        return (
            __unpickle_mutable_tenant_model_base,
            (model._for_tenant_model, tenant.natural_key(), model._meta.abstract)
        )
    return model.__name__

copyreg.pickle(MutableTenantModelBase, __pickle_mutable_tenant_model_base)


@receiver(mutable_class_prepared)
def contribute_to_related_mutable_class(sender, existing_model_class, **kwargs):
    """
    Since related fields are contributing to the related class only once (on
    the first time `class_prepared` is triggered) we make sure they also do
    the same with mutated classes they relate to in order to attach objects
    such as reverse descriptor.
    """
    for model in sender._meta.apps.get_models():
        for field in get_forward_fields(model._meta):
            remote_field = get_remote_field(field)
            if remote_field:
                remote_field_model = get_remote_field_model(field)
                # XXX: Remove the == conditional when dropping support for Django 1.8
                if remote_field_model is existing_model_class or remote_field_model == sender:
                    field.contribute_to_related_class(sender, remote_field)
                    set_remote_field_model(field, sender)


@receiver(pre_models_creation)
def manage_mutable_models(tenant, **kwargs):
    """
    Mark tenant mutable models as managed to prevent `create_tenant_schema`
    to create their associated table.
    """
    for model in tenant.models:
        if issubclass(model, MutableModel):
            model._meta.managed = True


@receiver(post_models_creation)
def unmanage_mutable_models(tenant, **kwargs):
    """
    Cleanup after our `manage_mutable_models` alteration.
    """
    for model in tenant.models:
        if issubclass(model, MutableModel):
            model._meta.managed = False


@receiver(pre_schema_deletion)
def cached_mutable_models(tenant, using, **kwargs):
    """
    Cache the mutable tenant model by bypassing their proxy and mark them as
    managed to prevent `drop_tenant_schema` from dropping their associated
    table twice.
    """
    models = []
    for model in tenant.models:
        if issubclass(model, MutableModel):
            # Access the underlying model class.
            model = model.__get__(None, None)
            opts = model._meta
            opts.managed = True
            # Repoint all local related object to the existing model class to
            # prevent access to the definition once it's deleted.
            for reverse_field in get_reverse_fields(opts):
                set_remote_field_model(reverse_field.field, model)
        models.append(model)
    tenant.models = tuple(models)



from __future__ import unicode_literals

from django.apps import AppConfig
from django.db.models import signals

from . import get_tenant_model


class TenancyConfig(AppConfig):
    name = 'tenancy'

    def clear_tenant_model_cache(self, **kwargs):
        get_tenant_model()._default_manager.clear_cache()

    def ready(self):
        # Prevents migrate from taking tenant models into consideration when
        # detecting changes.
        signals.pre_migrate.connect(self.clear_tenant_model_cache)



from __future__ import unicode_literals

import django

if django.VERSION >= (1, 9):
    def get_remote_field(field):
        return field.remote_field

    def get_remote_field_model(field):
        return field.remote_field.model

    def set_remote_field_model(field, model):
        field.remote_field.model = model

    def get_related_descriptor_field(descriptor):
        return descriptor.field

    from django.db.models.fields.related import lazy_related_operation  # noqa
else:
    def get_remote_field(field):
        return getattr(field, 'rel', None)

    def get_remote_field_model(field):
        return field.rel.to

    def set_remote_field_model(field, model):
        field.rel.to = model

    def get_related_descriptor_field(descriptor):
        return descriptor.related.field

    from django.db.models.fields.related import add_lazy_relation

    def lazy_related_operation(function, model, related_model, field):
        def operation(field, related, local):
            return function(local, related, field)
        add_lazy_relation(model, field, related_model, operation)



from __future__ import unicode_literals

import argparse

from django.contrib.auth.management.commands.createsuperuser import Command
from django.core.management.base import CommandError

from ... import get_tenant_model
from ...models import TenantModelBase


def get_tenant_by_natural_key(option, opt, value, parser):
    natural_key = value.split(',')
    tenant_model = get_tenant_model()
    tenant = tenant_model._default_manager.get_by_natural_key(*natural_key)
    parser.values.tenant = tenant


class TenantAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        tenant_model = get_tenant_model()
        tenant = tenant_model._default_manager.get_by_natural_key(*values)
        setattr(namespace, self.dest, tenant)


class Command(Command):
    help = 'Used to create a specific tenant superuser.'
    requires_system_checks = False

    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(*args, **kwargs)
        if not isinstance(self.UserModel, TenantModelBase):
            raise CommandError(
                "The defined user model (%s) is not tenant specific." % self.UserModel._meta
            )

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        parser.add_argument(
            'tenant', nargs='+', action=TenantAction,
            help='Specifies the tenant to use by natural key.'
        )

    def handle(self, *args, **options):
        tenant = options['tenant']
        self.UserModel = self.UserModel.for_tenant(tenant)
        return super(Command, self).handle(*args, **options)



from __future__ import unicode_literals

import logging

from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.core.management import call_command
from django.core.management.base import BaseCommand, CommandError
from django.utils.six.moves import input

from ... import get_tenant_model
from ...models import TenantModelBase


class CommandLoggingHandler(logging.StreamHandler):
    VERBOSITY_LEVELS = {
        0: logging.ERROR,
        1: logging.INFO,  # Would like to use WARNING here but 1 is the default
        2: logging.INFO,  # verbosity level and should map to INFO.
        3: logging.DEBUG,
    }

    def __init__(self, stdout, stderr, verbosity=1):
        self.error_stream = logging.StreamHandler(stderr)
        super(CommandLoggingHandler, self).__init__(stdout)
        self.setLevel(self.VERBOSITY_LEVELS[verbosity])

    def emit(self, record):
        if record.levelno >= logging.ERROR:
            self.error_stream.emit(record)
        else:
            super(CommandLoggingHandler, self).emit(record)


class Command(BaseCommand):
    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        parser.add_argument(
            '--noinput', '--no-input',
            action='store_false', dest='interactive', default=True,
            help='Tells Django to NOT prompt the user for input of any kind.'
        )
        parser.add_argument('fields', nargs='*')

    def handle(self, *args, **options):
        fields = options['fields']
        tenant_model = get_tenant_model()
        # Attempt to build the instance based on specified data
        try:
            tenant = tenant_model(None, *fields)
        except IndexError:
            opts = tenant_model._meta
            field_names = tuple(
                field.name for field in opts.local_fields if not field.primary_key
            )
            raise CommandError(
                "Number of args exceeds the number of fields for model %s.%s.\n"
                "Got %s when defined fields are %s." % (
                    opts.app_label,
                    opts.object_name,
                    fields,
                    field_names
                )
            )

        # Full clean the instance
        try:
            tenant.full_clean()
        except ValidationError as e:
            name, messages = e.message_dict.popitem()
            raise CommandError(
                'Invalid value for field "%s": %s.' % (name, messages[0])
            )

        # Redirect the output of the schema creation logger to our stdout.
        handler = CommandLoggingHandler(
            self.stdout._out, self.stderr._out, int(options['verbosity'])
        )
        logger = logging.getLogger('tenancy')
        logger.setLevel(handler.level)
        logger.addHandler(handler)

        # Create the tenant instance and create tables
        tenant.save(force_insert=True)

        # Remove the handler associated with the schema creation logger.
        logger.removeHandler(handler)
        logger.setLevel(logging.NOTSET)

        if options.get('interactive', True) and isinstance(get_user_model(), TenantModelBase):
            confirm = input(
                "\nYou just created a new tenant, which means you don't have "
                "any superusers defined.\nWould you like to create one "
                "now? (yes/no): "
            )
            while confirm not in ('yes', 'no'):
                confirm = input('Please enter either "yes" or "no": ')
            if confirm == 'yes':
                call_command('createtenantsuperuser', *tenant.natural_key(), **options)



from __future__ import unicode_literals

from contextlib import contextmanager
from itertools import chain

from django.apps import apps
from django.db import models
from django.utils.functional import cached_property

from .compat import get_remote_field, get_remote_field_model


def get_model(app_label, model_name):
    try:
        return apps.get_registered_model(app_label, model_name)
    except LookupError:
        pass


@contextmanager
def apps_lock():
    # The registry lock is not re-entrant so we must avoid acquiring it
    # during the initialization phase in order to prevent deadlocks.
    if apps.ready:
        with apps._lock:
            yield
    else:
        yield


def remove_from_app_cache(model_class, quiet=False):
    opts = model_class._meta
    apps = opts.apps
    app_label, model_name = opts.app_label, opts.model_name
    with apps_lock():
        try:
            model_class = apps.app_configs[app_label].models.pop(model_name)
        except KeyError:
            if not quiet:
                raise ValueError("%r is not cached" % model_class)
        apps.clear_cache()
        unreference_model(model_class)
    return model_class


def get_forward_fields(opts):
    return chain(
        opts.fields,
        opts.many_to_many
    )


def get_reverse_fields(opts):
    return opts._get_fields(forward=False, reverse=True, include_hidden=True)


def clear_opts_related_cache(model_class):
    opts = model_class._meta
    if not opts.apps.ready:
        return
    children = [
        related_object.related_model
        for related_object in opts.__dict__.get('related_objects', []) if related_object.parent_link
    ]
    opts._expire_cache()
    for child in children:
        clear_opts_related_cache(child)


def unreference_model(model):
    disconnect_signals(model)
    for field in get_forward_fields(model._meta):
        remote_field = get_remote_field(field)
        if field.model is model and remote_field:
            remote_field_model = get_remote_field_model(field)
            if isinstance(remote_field_model, models.base.ModelBase):
                clear_opts_related_cache(remote_field_model)
                rel_is_hidden = remote_field.is_hidden()
                # An accessor is added to related classes if they are not
                # hidden. However o2o fields *always* add an accessor
                # even if the relationship is hidden.
                o2o = isinstance(field, models.OneToOneField)
                if not rel_is_hidden or o2o:
                    try:
                        delattr(remote_field_model, remote_field.get_accessor_name())
                    except AttributeError:
                        pass


model_sender_signals = (
    models.signals.pre_init,
    models.signals.post_init,
    models.signals.pre_save,
    models.signals.post_save,
    models.signals.pre_delete,
    models.signals.post_delete,
    models.signals.m2m_changed,
)


def receivers_for_model(model):
    for signal in model_sender_signals:
        for receiver in signal._live_receivers(model):
            yield signal, receiver


def disconnect_signals(model):
    for signal, receiver in receivers_for_model(model):
        signal.disconnect(receiver, sender=model)


def clear_cached_properties(instance):
    """
    Clear the cache from the instance properties.
    """
    cls = type(instance)
    for attr in list(instance.__dict__):
        if isinstance(getattr(cls, attr, None), cached_property):
            instance.__dict__.pop(attr)


def get_postgresql_visible_constraints(cursor, table_name):
    """
    Retrieves any constraints or keys (unique, pk, fk, check, index) across one or more columns.
    """
    constraints = {}
    # Loop over the key table, collecting things as constraints
    # This will get PKs, FKs, and uniques, but not CHECK
    cursor.execute("""
        SELECT
            kc.constraint_name,
            kc.column_name,
            c.constraint_type,
            array(SELECT table_name::text || '.' || column_name::text
                  FROM information_schema.constraint_column_usage
                  WHERE constraint_name = kc.constraint_name)
        FROM information_schema.key_column_usage AS kc
        JOIN information_schema.table_constraints AS c ON
            kc.table_schema = c.table_schema AND
            kc.table_name = c.table_name AND
            kc.constraint_name = c.constraint_name
        WHERE
            kc.table_name = %s AND
            EXISTS(
                SELECT 1
                FROM pg_class AS c
                JOIN pg_namespace AS n ON
                    c.relnamespace = n.oid
                WHERE
                    n.nspname = kc.table_schema AND
                    c.relname = kc.table_name AND
                    pg_catalog.pg_table_is_visible(c.oid)
            )
        ORDER BY kc.ordinal_position ASC
    """, [table_name])
    for constraint, column, kind, used_cols in cursor.fetchall():
        # If we're the first column, make the record
        if constraint not in constraints:
            constraints[constraint] = {
                "columns": [],
                "primary_key": kind.lower() == "primary key",
                "unique": kind.lower() in ["primary key", "unique"],
                "foreign_key": tuple(used_cols[0].split(".", 1)) if kind.lower() == "foreign key" else None,
                "check": False,
                "index": False,
            }
        # Record the details
        constraints[constraint]['columns'].append(column)
    # Now get CHECK constraint columns
    cursor.execute("""
        SELECT kc.constraint_name, kc.column_name
        FROM information_schema.constraint_column_usage AS kc
        JOIN information_schema.table_constraints AS c ON
            kc.table_schema = c.table_schema AND
            kc.table_name = c.table_name AND
            kc.constraint_name = c.constraint_name
        WHERE
            c.constraint_type = 'CHECK' AND
            kc.table_name = %s AND
            EXISTS(
                SELECT 1
                FROM pg_class AS c
                JOIN pg_namespace AS n ON
                    c.relnamespace = n.oid
                WHERE
                    n.nspname = kc.table_schema AND
                    c.relname = kc.table_name AND
                    pg_catalog.pg_table_is_visible(c.oid)
            )
    """, [table_name])
    for constraint, column in cursor.fetchall():
        # If we're the first column, make the record
        if constraint not in constraints:
            constraints[constraint] = {
                "columns": [],
                "primary_key": False,
                "unique": False,
                "foreign_key": None,
                "check": True,
                "index": False,
            }
        # Record the details
        constraints[constraint]['columns'].append(column)
    # Now get indexes
    cursor.execute("""
        SELECT
            c2.relname,
            ARRAY(
                SELECT (SELECT attname FROM pg_catalog.pg_attribute WHERE attnum = i AND attrelid = c.oid)
                FROM unnest(idx.indkey) i
            ),
            idx.indisunique,
            idx.indisprimary
        FROM pg_catalog.pg_class c, pg_catalog.pg_class c2,
            pg_catalog.pg_index idx
        WHERE c.oid = idx.indrelid
            AND idx.indexrelid = c2.oid
            AND c.relname = %s
    """, [table_name])
    for index, columns, unique, primary in cursor.fetchall():
        if index not in constraints:
            constraints[index] = {
                "columns": list(columns),
                "primary_key": primary,
                "unique": unique,
                "foreign_key": None,
                "check": False,
                "index": True,
            }
    return constraints


@contextmanager
def patch_connection_introspection(connection):
    if connection.vendor == 'postgresql':
        get_constraints = connection.introspection.get_constraints
        connection.introspection.get_constraints = get_postgresql_visible_constraints
    try:
        yield
    finally:
        if connection.vendor == 'postgresql':
            connection.introspection.get_constraints = get_constraints



#!/usr/bin/env python
from __future__ import unicode_literals

from setuptools import setup, find_packages

from tenancy import __version__


setup(
    name='django-tenancy',
    version=__version__,
    description='Handle multi-tenancy in Django with no additional global state using schemas.',
    url='https://github.com/charettes/django-tenancy',
    author='Simon Charette',
    author_email='charette.s+pypi@gmail.com',
    install_requires=[
        'Django>=1.8',
    ],
    packages=find_packages(exclude=['tests', 'tests.*']),
    license='MIT License',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    extras_require={
        'hosts': ['django-hosts'],
        'mutant': ['django-mutant>=0.2.1'],
    }
)



from __future__ import unicode_literals

from contextlib import contextmanager

from django.apps import apps
from django.db.backends.utils import truncate_name
from django.db.migrations import operations
from django.db.migrations.operations.base import Operation
from django.utils.six import iteritems

from .models import Managed, db_schema_table
from .utils import patch_connection_introspection


class TenantOperation(Operation):
    def get_tenant_model(self, app_label, from_state, to_state):
        raise NotImplementedError

    @contextmanager
    def tenant_context(self, tenant, schema_editor):
        connection = schema_editor.connection
        cursor = connection.cursor()
        if connection.vendor == 'postgresql':
            sql = "SET search_path = %s, public" % tenant.db_schema
            cursor.execute(sql)
            schema_editor.deferred_sql.append(sql)
        with patch_connection_introspection(connection):
            yield
        if connection.vendor == 'postgresql':
            sql = 'RESET search_path'
            cursor.execute(sql)
            schema_editor.deferred_sql.append(sql)

    def create_tenant_project_state(self, tenant, state, connection):
        managed = Managed("%s.%s" % (tenant._meta.app_label, tenant._meta.object_name))
        project_state = state.clone()
        for (app_label, model_name), model_state in iteritems(project_state.models):
            options = model_state.options
            if options.get('managed') == managed:
                db_table = options.get('db_table')
                if not db_table:
                    db_table = truncate_name("%s_%s" % (app_label, model_name), connection.ops.max_name_length())
                if not connection.vendor == 'postgresql':
                    db_table = db_schema_table(tenant, db_table)
                options.update(
                    managed=True,
                    db_table=db_table,
                )
            project_state.reload_model(app_label, model_name)
        return project_state

    def tenant_operation(self, tenant_model, operation, app_label, schema_editor, from_state, to_state):
        connection = schema_editor.connection
        for tenant in tenant_model._base_manager.all():
            tenant_from_state = self.create_tenant_project_state(tenant, from_state, connection)
            tenant_to_state = self.create_tenant_project_state(tenant, to_state, connection)
            with self.tenant_context(tenant, schema_editor):
                operation(app_label, schema_editor, tenant_from_state, tenant_to_state)

    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        tenant_model = self.get_tenant_model(app_label, from_state, to_state)
        operation = super(TenantOperation, self).database_forwards
        self.tenant_operation(tenant_model, operation, app_label, schema_editor, from_state, to_state)

    def database_backwards(self, app_label, schema_editor, from_state, to_state):
        tenant_model = self.get_tenant_model(app_label, to_state, from_state)
        operation = super(TenantOperation, self).database_backwards
        self.tenant_operation(tenant_model, operation, app_label, schema_editor, from_state, to_state)


class TenantModelOperation(TenantOperation):
    def get_operation_model_state(self, app_label, from_state, to_state):
        return to_state.models[app_label, self.name_lower]

    def get_tenant_model(self, app_label, from_state, to_state):
        model_state = self.get_operation_model_state(app_label, from_state, to_state)
        managed = model_state.options.get('managed')
        return apps.get_model(managed.tenant_model)


class CreateModel(TenantModelOperation, operations.CreateModel):
    pass


class DeleteModel(TenantModelOperation, operations.DeleteModel):
    def get_operation_model_state(self, app_label, from_state, to_state):
        return from_state.models[app_label, self.name_lower]


class RenameModel(TenantModelOperation, operations.RenameModel):
    def get_operation_model_state(self, app_label, from_state, to_state):
        return from_state.models[app_label, self.old_name_lower]

    database_backwards = operations.RenameModel.database_backwards


class AlterModelTable(TenantModelOperation, operations.AlterModelTable):
    database_backwards = operations.AlterModelTable.database_backwards


class AlterUniqueTogether(TenantModelOperation, operations.AlterUniqueTogether):
    database_backwards = operations.AlterUniqueTogether.database_backwards


class AlterIndexTogether(TenantModelOperation, operations.AlterIndexTogether):
    database_backwards = operations.AlterIndexTogether.database_backwards


class TenantModelFieldOperation(TenantModelOperation):
    def get_operation_model_state(self, app_label, from_state, to_state):
        return from_state.models[app_label, self.model_name_lower]


class AddField(TenantModelFieldOperation, operations.AddField):
    pass


class RemoveField(TenantModelFieldOperation, operations.RemoveField):
    pass


class AlterField(TenantModelFieldOperation, operations.AlterField):
    database_backwards = operations.AlterField.database_backwards


class RenameField(TenantModelFieldOperation, operations.RenameField):
    pass


class TenantSpecialOperation(TenantOperation):
    def __init__(self, tenant_model, *args, **kwargs):
        self.tenant_model = tenant_model
        super(TenantSpecialOperation, self).__init__(*args, **kwargs)

    def get_tenant_model(self, app_label, from_state, to_state):
        return self.tenant_model


class RunPython(TenantSpecialOperation, operations.RunPython):
    pass


class RunSQL(TenantSpecialOperation, operations.RunSQL):
    pass



from __future__ import unicode_literals

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import DEFAULT_DB_ALIAS, connections
from django.http import Http404

from . import get_tenant_model
from .settings import HOST_NAME


class TenantHostMiddleware(object):
    def __init__(self):
        try:
            import django_hosts  # NOQA
        except ImportError:
            raise ImproperlyConfigured(
                'You must install django-hosts in order to '
                'use `TenantHostMiddleware`.'
            )
        path = "%s.%s" % (self.__module__, self.__class__.__name__)
        for middleware in settings.MIDDLEWARE_CLASSES:
            if middleware == 'django_hosts.middleware.HostsRequestMiddleware':
                break
            elif middleware == path:
                raise ImproperlyConfigured(
                    "Make sure that 'django_hosts.middleware.HostsRequestMiddleware' "
                    "is placed before '%s' in your `MIDDLEWARE_CLASSES` "
                    "setting." % path
                )
        self.tenant_model = get_tenant_model()
        self.attr_name = self.tenant_model.ATTR_NAME

    def process_request(self, request):
        if request.host.name == HOST_NAME:
            match = request.host.compiled_regex.match(request.get_host())
            lookups = match.groupdict()
            tenant_model = self.tenant_model
            try:
                tenant = tenant_model._default_manager.get(**lookups)
            except tenant_model.DoesNotExist:
                raise Http404(
                    "No tenant found for specified lookups: %r" % lookups
                )
            setattr(request, self.attr_name, tenant)


class GlobalTenantMiddleware(object):
    """
    Middleware that assigns the request's tenant attribute to the default
    connection object. This unfortunate global state is required in order to
    allow things such as a tenant custom user with the required auth backend.
    """

    def __init__(self):
        self.attr_name = get_tenant_model().ATTR_NAME

    def get_global_state(self):
        return connections[DEFAULT_DB_ALIAS]

    def pollute_global_state(self, tenant):
        setattr(self.get_global_state(), self.attr_name, tenant)

    def clean_global_state(self):
        global_state = self.get_global_state()
        if hasattr(global_state, self.attr_name):
            delattr(global_state, self.attr_name)

    def process_request(self, request):
        self.pollute_global_state(
            getattr(request, self.attr_name, None)
        )

    def process_response(self, request, response):
        self.clean_global_state()
        return response

    def process_exception(self, request, exception):
        self.clean_global_state()
